import tensorflow as tf
import numpy as np


@tf.function
def intersectSegment(a: tf.Tensor, b: tf.Tensor, returnMask=False):
    """Intersects list of segments

    Literally taken from https://en.wikipedia.org/wiki/Line%E2%80%93line_intersection
    Args:
        a (tf.Tensor): (N,2,2) endpoint of line
        b (tf.Tensor): (N,2,2) endpoint of line

    Returns:
        tf.RaggedTensor: (N,None,2) of the intersections
    """
    x1 = a[..., 0, 0]
    y1 = a[..., 0, 1]

    x2 = a[..., 1, 0]
    y2 = a[..., 1, 1]

    x3 = b[..., 0, 0]
    y3 = b[..., 0, 1]

    x4 = b[..., 1, 0]
    y4 = b[..., 1, 1]

    D = (x1 - x2) * (y3 - y4) - (y1 - y2) * (x3 - x4)
    t = tf.math.divide_no_nan((x1 - x3) * (y3 - y4) - (y1 - y3) * (x3 - x4), D)
    u = tf.math.divide_no_nan((x1 - x3) * (y1 - y2) - (y1 - y3) * (x1 - x2), D)

    inT = tf.math.logical_and(t >= 0, t <= 1)
    inU = tf.math.logical_and(u >= 0, u <= 1)
    inBoth = tf.math.logical_and(inT, inU)
    mask = tf.logical_and(tf.math.abs(D) > 1e-8, inBoth)
    res = a[..., 0, :] + t[..., None] * (a[..., 1, :] - a[..., 0, :])
    if returnMask:
        return res, mask
    return tf.RaggedTensor.from_row_lengths(tf.boolean_mask(res, mask),
                                            tf.cast(mask, tf.int64))


@tf.function
def pointsInPolygon(polygon, points):
    """
    Test if a point lies in a given polygon

    Args:
        polygon (tf.Tensor): (N,4,2) ragged tensor of vertices of the polygons (order does matter)
        point (tf.Tensor): (N,M,2) points to test for

    Returns:
        tf.Tensor: (N,M) boolean vector True if points lies inside P
    """
    polygonRolled = tf.roll(polygon, -1, axis=-2)
    side_criterion = (polygonRolled[..., 0] - polygon[..., 0])[:, None, :] * (
        points[..., 1][..., None] - polygon[..., 1][:, None, :]) - (
            points[..., 0][..., None] - polygon[..., 0][:, None, :]) * (
                polygonRolled[..., 1] - polygon[..., 1])[:, None, :]
    onUpEdge = tf.math.logical_and(
        points[..., 1][..., None] >= polygon[..., 1][:, None, :],
        points[..., 1][..., None] < polygonRolled[..., 1][:, None, :],
    )

    onDownEdge = tf.math.logical_and(
        points[..., 1][..., None] < polygon[..., 1][:, None, :],
        points[..., 1][..., None] >= polygonRolled[..., 1][:, None, :],
    )
    windingNumber = tf.cast(
        tf.math.logical_and(onUpEdge, side_criterion > 0.0),
        tf.int8) - tf.cast(
            tf.math.logical_and(onDownEdge, side_criterion < 0.0), tf.int8)

    windingNumber = tf.math.reduce_sum(windingNumber, axis=-1)
    return windingNumber != 0


def _edges(a):
    return tf.concat(
        [
            a[:, :, None, :],
            tf.roll(a, -1, axis=-2)[:, :, None, :],
        ],
        axis=-2,
    )


@tf.function
def intersectPolygons(a, b):
    """
    Computes the intersection of polygons. These polygons must have
    the same number of vertices.

    Args:
        a (tf.Tensor): (N,K,2) array of polygons to compute intersection
        b (tf.Tensor): (N,K,2) array of polygons to compute intersection

    Returns:
        tf.Tensor: (N,L,2) array of the intersection definition.  It
            may contains duplicate points. L can grow up to K ** 2 + 2
            * K, depending on the actual intersections and if there
            are some duplicate created by the algorithm.

    """
    nbPolygon = tf.shape(a)[0]
    nbVertices = tf.shape(a)[1]

    # 1. collects vertices of each polygon that lies in the other polygon.
    cornersAInB = pointsInPolygon(b, a)
    cornersBInA = pointsInPolygon(a, b)
    corners = tf.concat([a, b], axis=1)
    maskInside = tf.concat([cornersAInB, cornersBInA], axis=1)

    # 2. Collects all edges intersections
    edgesA = tf.reshape(
        tf.tile(_edges(a), multiples=[1, 1, nbVertices, 1]),
        shape=[-1, nbVertices**2, 2, 2],
    )
    edgesB = tf.tile(_edges(b), multiples=[1, nbVertices, 1, 1])  # N,K*K,2,2
    intersections, maskIntersections = intersectSegment(edgesA,
                                                        edgesB,
                                                        returnMask=True)

    mask = tf.concat([maskInside, maskIntersections], axis=1)
    sizes = tf.math.reduce_sum(tf.cast(mask, tf.int64), axis=-1)
    allCorners = tf.RaggedTensor.from_row_lengths(
        tf.boolean_mask(tf.concat([corners, intersections], axis=1), mask),
        sizes,
    )

    # 3. Order all collected vertices in trigonometric order arround
    # the centroid. Note: it may not be the exact centroid due to
    # duplicate, but since its a convex shape it won't affect the
    # order, as we only need to oder them arround a point inside the
    # polygon.

    cornersAtCentroid = allCorners - tf.math.reduce_mean(allCorners,
                                                         axis=1)[:, None, :]
    angles = tf.math.atan2(
        cornersAtCentroid[..., 1],
        cornersAtCentroid[..., 0],
    )

    indexes = tf.argsort(angles.to_tensor(default_value=float("Inf")), axis=1)
    sortedCorners = tf.gather(allCorners.to_tensor(), indexes, batch_dims=1)
    finalMask = tf.RaggedTensor.from_row_lengths(
        tf.ones(tf.math.reduce_sum(sizes), tf.bool),
        sizes,
    )

    # Now we repeat the first vertices in the last indices in order
    # for the result to be a valid polygon for polygonArea function.
    return tf.where(
        finalMask.to_tensor()[..., None],
        sortedCorners,
        sortedCorners[:, 0, :][:, None, :],
    )


def uniqueVertex(a):
    res = []
    for i in range(a.shape[0]):
        p = a[i, ...].numpy()
        indexes = np.unique(p, axis=0, return_index=True)[1]
        res.append(np.array([p[index, ...] for index in sorted(indexes)]))
    return tf.ragged.constant(res)


@tf.function
def polygonArea(a):
    """
    Computes the area of a series of polygons. Polygons must have the
    same number of vertices. Different number of vertices can be faked
    by repeating any number of vertices consecutively. (Order does
    matter).

    Args:
        a (tf.Tensor): (N,M,2) N polygon of M vertices

    Returns:
        tf.Tensor: (N,) The area of the polygons

    """
    rolledA = tf.roll(a, shift=-1, axis=-2)
    return 0.5 * tf.abs(
        tf.math.reduce_sum(
            a[..., 0] * rolledA[..., 1] - a[..., 1] * rolledA[..., 0],
            axis=-1))


@tf.function
def IoUMatrix(a, b):
    """
    Computes Intersection over union matrix of collection of convex polygons.

    Args:
        a (tf.Tensor): (N,K,2) a list of convex polygons.
        b (tf.Tensor): (M,K,2) a list of convex polygons.

    Returns:
        tf.Tensor: (N,M) the matrices of iou of a and b.
    """
    numA = tf.shape(a)[0]
    numB = tf.shape(b)[0]
    numVertices = tf.shape(a)[1]
    flatA = tf.reshape(tf.tile(a, multiples=[1, numB, 1]),
                       shape=[-1, numVertices, 2])
    flatB = tf.tile(b, multiples=[numA, 1, 1])

    intersections = intersectPolygons(flatA, flatB)
    areaA = polygonArea(flatA)
    areaB = polygonArea(flatB)
    areaIntersection = polygonArea(intersections)
    return tf.reshape(areaIntersection / (areaA + areaB - areaIntersection),
                      shape=[numA, numB])



import unittest
import tf_convex_polygon_iou._data as udata
import tf_convex_polygon_iou._tf_implementation as tfi
import numpy as np


class TensorflowTestCase(unittest.TestCase):

    def test_segment_intersection(self):
        for d in udata.SEGMENT_INTERSECTIONS:
            with self.subTest(str(d)):
                points = tfi.intersectSegment(
                    d.Segment1[None, ...],
                    d.Segment2[None, ...],
                )
                np.testing.assert_almost_equal(
                    points[0, ...].numpy(),
                    d.Expected,
                )

    def test_segment_intersection_ragged(self):
        segment1, segment2, expected = udata.SegmentIntersectionData.ragged(
            udata.SEGMENT_INTERSECTIONS)
        r = tfi.intersectSegment(segment1, segment2)
        np.testing.assert_almost_equal(
            r.to_tensor(default_value=np.Inf).numpy(),
            expected.to_tensor(default_value=np.Inf))

    def test_points_in_polygons(self):
        for d in udata.POINTS_IN_POLYGONS:
            with self.subTest(str(d)):
                np.testing.assert_almost_equal(
                    tfi.pointsInPolygon(d.Polygon[None, ...],
                                        d.Points[None, ...])[0, ...].numpy(),
                    d.Expected,
                )

    def test_points_in_polygons_ragged(self):
        polygons, points, expected = udata.PointsInPolygonData.ragged(
            udata.POINTS_IN_POLYGONS)
        r = tfi.pointsInPolygon(polygons, points)
        np.testing.assert_almost_equal(r.numpy(), expected.numpy())

    def test_polygons_intersection(self):
        for d in udata.POLYGON_INTERSECTIONS:
            with self.subTest(str(d)):
                np.testing.assert_almost_equal(
                    tfi.uniqueVertex(
                        tfi.intersectPolygons(
                            d.Polygon1[None, ...],
                            d.Polygon2[None, ...],
                        ))[0, ...].numpy(), d.Expected)

    def test_polygons_intersection_ragged(self):
        polygon1, polygon2, expected = udata.PolygonsIntersectionData.ragged(
            udata.POLYGON_INTERSECTIONS)

        r = tfi.intersectPolygons(polygon1, polygon2)
        r = tfi.uniqueVertex(r)
        np.testing.assert_almost_equal(
            r.to_tensor(default_value=float('Inf')).numpy(),
            expected.to_tensor(default_value=float('Inf')).numpy())

    def test_polygon_area(self):
        for d in udata.POLYGON_AREA:
            with self.subTest(str(d)):
                np.testing.assert_almost_equal(tfi.polygonArea(d.Polygon),
                                               d.Area)

    def test_polygon_area_ragged(self):
        polygons, area = udata.PolygonArea.ragged(udata.POLYGON_AREA)
        np.testing.assert_almost_equal(
            tfi.polygonArea(polygons).numpy(), area.numpy())

    def test_quad_IoU(self):
        for d in udata.POLYGON_INTERSECTIONS:
            with self.subTest(str(d)):
                np.testing.assert_almost_equal(
                    tfi.IoUMatrix(d.Polygon1[None, ...],
                                  d.Polygon2[None, ...])[0, 0], d.IoU)

    def test_quad_IoUMatrix(self):
        polygons1, polygons2, expected = udata.PolygonsIntersectionData.IoUMatrix(
            udata.POLYGON_INTERSECTIONS)
        np.testing.assert_almost_equal(
            tfi.IoUMatrix(polygons1, polygons2).numpy(), expected.numpy())



import unittest
import tf_convex_polygon_iou._data as udata
import tf_convex_polygon_iou._numpy_implementation as npi
import numpy as np


class NumpyTestCase(unittest.TestCase):

    def test_segment_intersection(self):
        for d in udata.SEGMENT_INTERSECTIONS:
            with self.subTest(str(d)):
                np.testing.assert_almost_equal(
                    npi.intersectSegment(
                        d.Segment1,
                        d.Segment2,
                    ),
                    d.Expected,
                )

    def test_points_in_polygon(self):
        for d in udata.POINTS_IN_POLYGONS:
            with self.subTest(str(d)):
                np.testing.assert_almost_equal(
                    npi.pointsInPolygon(d.Polygon, d.Points),
                    d.Expected,
                )

    def test_polygons_intersection(self):
        for d in udata.POLYGON_INTERSECTIONS:
            with self.subTest(str(d)):
                np.testing.assert_almost_equal(
                    npi.uniqueVertex(
                        npi.intersectPolygons(
                            d.Polygon1,
                            d.Polygon2,
                        )), d.Expected)

    def test_polygon_area(self):
        for d in udata.POLYGON_AREA:
            with self.subTest(str(d)):
                np.testing.assert_almost_equal(npi.polygonArea(d.Polygon),
                                               d.Area)

    def test_IoU(self):
        for d in udata.POLYGON_INTERSECTIONS:
            with self.subTest(str(d)):
                np.testing.assert_almost_equal(
                    npi.IoUMatrix(d.Polygon1[None, ...],
                                  d.Polygon2[None, ...])[0, 0], d.IoU)

    def test_IoUMatrix(self):
        polygons1, polygons2, expected = udata.PolygonsIntersectionData.IoUMatrix(
            udata.POLYGON_INTERSECTIONS)
        np.testing.assert_almost_equal(
            npi.IoUMatrix(polygons1.numpy(), polygons2.numpy()),
            expected.numpy())



import numpy as np


def intersectSegment(a: np.ndarray, b: np.ndarray):
    """Intersects two segments

    Literally taken from https://en.wikipedia.org/wiki/Line%E2%80%93line_intersection
    Args:
        a (np.ndarray): (2,2) endpoint of line
        b (np.ndarray): (2,2) endpoint of line

    Returns:
        np.ndarray: (1,2) or (1,0) the intersection point if it exists
    """
    x1, y1 = a[0, :]
    x2, y2 = a[1, :]

    x3, y3 = b[0, :]
    x4, y4 = b[1, :]

    D = (x1 - x2) * (y3 - y4) - (y1 - y2) * (x3 - x4)
    if abs(D) < 1e-8:
        return np.zeros((0, 2))

    t = (x1 - x3) * (y3 - y4) - (y1 - y3) * (x3 - x4)
    t /= D

    u = (x1 - x3) * (y1 - y2) - (y1 - y3) * (x1 - x2)
    u /= D

    if t < 0 or t > 1 or u < 0 or u > 1:
        return np.zeros((0, 2))

    return np.array([[x1 + t * (x2 - x1), y1 + t * (y2 - y1)]])


def pointsInPolygon(polygon: np.ndarray, points: np.ndarray):
    """Test if a point lies in a given polygon

    Args:
        polygon (np.ndarray): (K,2) vertices of the polygon (order does matter)
        point (np.ndarray): (N,2) points to test for

    Returns:
        np.ndarray: (N,) boolean vector True if points lies inside P
    """
    polygonRolled = np.roll(polygon, -1, axis=0)

    side_criterion = (polygonRolled[:, 0] - polygon[:, 0]) * (
        points[:, 1][:, None] - polygon[:, 1]) - (
            points[:, 0][:, None] - polygon[:, 0]) * (polygonRolled[:, 1] -
                                                      polygon[:, 1])
    onUpEdge = (points[:, 1][:, None] >=
                polygon[:, 1]) * (points[:, 1][:, None] < polygonRolled[:, 1])
    onDownEdge = (points[:, 1][:, None] < polygon[:, 1]) * (
        points[:, 1][:, None] >= polygonRolled[:, 1])

    windingNumber = (onUpEdge * (side_criterion > 0.0)).astype(
        np.int32) - (onDownEdge * (side_criterion < 0.0)).astype(np.int32)
    windingNumber = np.sum(windingNumber, axis=-1)
    return windingNumber != 0


def intersectPolygons(a: np.ndarray, b: np.ndarray):
    """
    Computes the intersection of quads

    Args:
        a (np.ndarray): (K,2) array of quad to compute intersection
        b (np.ndarray): (K,2) array of quad to compute intersection

    Returns:
        np.ndarray: (N,2) array of the intersections it may contains duplicate
            points. N can go up to K ** 2 + 2 * K.
    """
    cornersAinB = pointsInPolygon(b, a)
    cornersBinA = pointsInPolygon(a, b)

    corners_inside = np.concatenate([a[cornersAinB, :], b[cornersBinA, :]],
                                    axis=0)
    intersections = np.zeros((16, 2))
    numberOfIntersections = 0
    for startA, endA in zip(a, np.roll(a, -1, axis=0)):
        for startB, endB in zip(b, np.roll(b, -1, axis=0)):
            intersection = intersectSegment(
                np.concatenate([endA[None, :], startA[None, :]], axis=0),
                np.concatenate([endB[None, :], startB[None, :]], axis=0),
            )
            if intersection.shape[0] == 1:
                intersections[numberOfIntersections] = intersection
                numberOfIntersections += 1
    allCorners = np.concatenate(
        [corners_inside, intersections[:numberOfIntersections, :]], axis=0)
    if allCorners.shape[0] == 0:
        return np.array([], dtype=np.float32)
    cornersAtCentroid = allCorners - np.mean(allCorners, axis=0)
    angles = np.arctan2(cornersAtCentroid[:, 1], cornersAtCentroid[:, 0])
    return allCorners[np.argsort(angles), :]


def polygonArea(a: np.ndarray):
    """
    Computes the area of a polygon

    Args:
        a (np.ndarray): (N,2)
    Returns:
        float32: the area of the polygon
    """
    if len(a.shape) < 2 or a.shape[-2] == 0:
        return 0.0
    rolledA = np.roll(a, -1, axis=-2)
    return 0.5 * abs(np.sum(a[:, 0] * rolledA[:, 1] - a[:, 1] * rolledA[:, 0]))


def uniqueVertex(a: np.ndarray):
    """Returns unique polygon vertices, keeping the order

    Args:
        a (np.ndarray): (N,2) a list of vertices

    Returns:
        np.ndarray: unique vertices in a, keeping it sorted
    """
    indexes = np.unique(a, axis=0, return_index=True)[1]
    return np.array([a[index, ...] for index in sorted(indexes)])


def IoUMatrix(a: np.ndarray, b: np.ndarray):
    """Computes Intersection over union matrix of collection of boxes.

    Args:
        a (np.ndarray): (N,4,2) a list of convex quads
        b (np.ndarray): (M,4,2) a list of convex quads

    Returns:
        np.ndarray: (N,M) the matrices of iou of a and b
    """
    res = np.zeros((a.shape[0], b.shape[0]))
    for i in range(a.shape[0]):
        for j in range(b.shape[0]):
            areaA = polygonArea(a[i, ...])
            areaB = polygonArea(b[j, ...])
            inter = intersectPolygons(a[i, ...], b[j, ...])
            areaInter = polygonArea(inter)
            res[i, j] = areaInter / (areaA + areaB - areaInter)
    return res



from dataclasses import dataclass
import numpy as np
import math
import tensorflow as tf

sqrt2 = math.sqrt(2)

@dataclass
class SegmentIntersectionData:
    Segment1: np.ndarray
    Segment2: np.ndarray
    Expected: np.ndarray

    @staticmethod
    def ragged(l):
        return (
            tf.concat([d.Segment1[None, :, :] for d in l], axis=0),
            tf.concat([d.Segment2[None, :, :] for d in l], axis=0),
            tf.ragged.constant([d.Expected for d in l]),
        )


SEGMENT_INTERSECTIONS = [
    SegmentIntersectionData(
        Segment1=np.array([[0, 0], [0, 1]], dtype=np.float32),
        Segment2=np.array([[0, 0], [1, 0]], dtype=np.float32),
        Expected=np.array([[0, 0]], dtype=np.float32),
    ),
    SegmentIntersectionData(
        Segment1=np.array([[0, 0], [1, 1]], dtype=np.float32),
        Segment2=np.array([[1, 0], [0, 1]], dtype=np.float32),
        Expected=np.array([[0.5, 0.5]], dtype=np.float32),
    ),
    # co-linear condition
    SegmentIntersectionData(
        Segment1=np.array([[0, 0], [0, 1]], dtype=np.float32),
        Segment2=np.array([[0, 0], [0, 2]], dtype=np.float32),
        Expected=np.zeros((0, 2), dtype=np.float32),
    ),
    SegmentIntersectionData(
        Segment1=np.array([[0, 0], [0, 1]], dtype=np.float32),
        Segment2=np.array([[1, 0], [1, 1]], dtype=np.float32),
        Expected=np.zeros(shape=(0, 2), dtype=np.float32),
    ),
    # intersection outside of segment
    SegmentIntersectionData(
        Segment1=np.array([[0, 0], [0, 1]], dtype=np.float32),
        Segment2=np.array([[1, 1], [2, 1]], dtype=np.float32),
        Expected=np.zeros(shape=(0, 2), dtype=np.float32),
    ),
    SegmentIntersectionData(
        Segment1=np.array([[0, 0], [0, 1]], dtype=np.float32),
        Segment2=np.array([[3, 2], [2, 2]], dtype=np.float32),
        Expected=np.zeros(shape=(0, 2), dtype=np.float32),
    ),
]


@dataclass
class PointsInPolygonData:
    Polygon: np.ndarray
    Points: np.ndarray
    Expected: np.ndarray

    @staticmethod
    def ragged(l):
        return (
            tf.constant([d.Polygon for d in l]),
            tf.constant([d.Points for d in l]),
            tf.constant([d.Expected for d in l]),
        )


POINTS_IN_POLYGONS = [
    # -- With a square --o
    PointsInPolygonData(
        np.array([[-1, -1], [1, -1], [1, 1], [-1, 1]], dtype=np.float32),
        np.array(
            [
                [0, 0],
                [1, 2],
                [1, -2],
                [-1, -1],
                [1, 1],
                [-1, 1],
                [1, -1],
                [0, 1 - 1e-6],
                [0, -1 + 1e-6],
                [-1 + 1e-6, 0],
                [1 - 1e-6, 0],
                [1, 0],
                [0, 1],
                [-1, 0],
                [0, -1],
            ],
            dtype=np.float32,
        ),
        np.array(
            [
                True,
                False,
                False,
                True,
                False,
                False,
                False,
                True,
                True,
                True,
                True,
                False,
                False,
                True,
                True,
            ],
            dtype=bool,
        ),
    ),
    # -- With an hourglass --
    # just above the center is inside
    PointsInPolygonData(
        np.array([[-1, -1], [1, -1], [-1, 1], [1, 1]], dtype=np.float32),
        np.array(
            [
                [0.0, 1.0e-6],
                [1.0e-6, 0.0],
                [0.0, 0.0],
                [-1, -1],
                [1, 1],
                [-1, 1],
                [1, -1],
                [0, 1 - 1e-6],
                [0, -1 + 1e-6],
                [-1 + 1e-6, 0],
                [1 - 1e-6, 0],
                [1, 0],
                [0, 1],
                [-1, 0],
                [0, -1],
            ],
            dtype=np.float32,
        ),
        np.array(
            [
                True,
                False,
                False,
                True,
                False,
                False,
                False,
                True,
                True,
                False,
                False,
                False,
                False,
                False,
                True,
            ],
            dtype=bool,
        ),
    ),
    PointsInPolygonData(
        np.array([[0, -1], [1, 0], [0, 1], [-1, 0]], dtype=np.float32),
        np.array(
            [
                [0.0, 0.0],
                [-1.0, 1.0],
                [1.0, 1.0],
                [-1, 0],
                [0, -1],
                [1, 0],
                [0, 1],
                [0, 1 - 1e-6],
                [0, -1 + 1e-6],
                [-1 + 1e-6, 0],
                [1 - 1e-6, 0],
                [sqrt2 / 2, sqrt2 / 2],
                [-sqrt2 / 2, sqrt2 / 2],
                [sqrt2 / 2, -sqrt2 / 2],
                [-sqrt2 / 2, -sqrt2 / 2],
            ],
            dtype=np.float32,
        ),
        np.array(
            [
                True,
                False,
                False,
                True,
                False,
                False,
                False,
                True,
                True,
                True,
                True,
                False,
                False,
                False,
                False,
            ],
            dtype=bool,
        ),
    ),
]


@dataclass
class PolygonsIntersectionData:
    Polygon1: np.ndarray
    Polygon2: np.ndarray
    Expected: np.ndarray
    IoU: float

    @staticmethod
    def ragged(l):
        return (
            tf.constant([d.Polygon1 for d in l]),
            tf.constant([d.Polygon2 for d in l]),
            tf.ragged.constant([d.Expected for d in l]),
        )

    @staticmethod
    def IoUMatrix(l):
        defaultPolygon = l[0].Polygon1
        filtered = [d for d in l if (d.Polygon1 == defaultPolygon).all()]
        return (
            tf.constant([d.Polygon2 for d in filtered]),
            tf.concat([defaultPolygon[None, ...], [[[0, 0]] * 4]], axis=0),
            tf.concat(
                [[[d.IoU] for d in filtered], [[0.0]] * len(filtered)],
                axis=-1,
            ),
        )


POLYGON_INTERSECTIONS = [
    PolygonsIntersectionData(
        Polygon1=np.array(
            [
                [-1, -1],
                [-1, 1],
                [1, 1],
                [1, -1],
            ],
            dtype=np.float32,
        ),
        Polygon2=np.array(
            [
                [-1, -1],
                [-1, 1],
                [1, 1],
                [1, -1],
            ],
            dtype=np.float32,
        ),
        Expected=np.array(
            [
                [-1, -1],
                [1, -1],
                [1, 1],
                [-1, 1],
            ],
            dtype=np.float32,
        ),
        IoU=1.0,
    ),
    PolygonsIntersectionData(
        Polygon1=np.array(
            [
                [-1, -1],
                [-1, 1],
                [1, 1],
                [1, -1],
            ],
            dtype=np.float32,
        ),
        Polygon2=np.array(
            [
                [0, 0],
                [0, 2],
                [2, 2],
                [2, 0],
            ],
            dtype=np.float32,
        ),
        Expected=np.array(
            [
                [0, 0],
                [1, 0],
                [1, 1],
                [0, 1],
            ],
            dtype=np.float32,
        ),
        IoU=1 / 7,
    ),
    PolygonsIntersectionData(
        Polygon1=np.array(
            [
                [-1, -1],
                [-1, 1],
                [1, 1],
                [1, -1],
            ],
            dtype=np.float32,
        ),
        Polygon2=np.array(
            [
                [0, sqrt2],
                [sqrt2, 0],
                [0, -sqrt2],
                [-sqrt2, 0],
            ],
            dtype=np.float32,
        ),
        Expected=np.array(
            [
                [-1, -sqrt2 + 1],
                [-sqrt2 + 1, -1],
                [sqrt2 - 1, -1],
                [1, -sqrt2 + 1],
                [1, sqrt2 - 1],
                [sqrt2 - 1, 1],
                [-sqrt2 + 1, 1],
                [-1, sqrt2 - 1],
            ],
            dtype=np.float32,
        ),
        IoU=(1 - (sqrt2 - 1) ** 2) / (1 + (sqrt2 - 1) ** 2),
    ),
    PolygonsIntersectionData(
        Polygon1=np.array(
            [
                [-1, -1],
                [-1, 1],
                [1, 1],
                [1, -1],
            ],
            dtype=np.float32,
        ),
        Polygon2=np.array(
            [
                [-1, 1],
                [0, 0],
                [1, 0],
                [2, 1],
            ],
            dtype=np.float32,
        ),
        Expected=np.array(
            [
                [0, 0],
                [1, 0],
                [1, 1],
                [-1, 1],
            ],
            dtype=np.float32,
        ),
        IoU=(2 - 1 / 2) / (4 + 1 / 2),
    ),
    PolygonsIntersectionData(
        Polygon1=np.array(
            [
                [-1, -1],
                [-1, 1],
                [1, 1],
                [1, -1],
            ],
            dtype=np.float32,
        ),
        Polygon2=np.array(
            [
                [1, 1],
                [2, 1],
                [2, 2],
                [1, 2],
            ],
            dtype=np.float32,
        ),
        Expected=np.array(
            [
                [1, 1],
            ],
            dtype=np.float32,
        ),
        IoU=0.0,
    ),
    PolygonsIntersectionData(
        Polygon1=np.array(
            [
                [0, 0],
                [8, 0],
                [8, 10],
                [0, 10],
            ],
            dtype=np.float32,
        ),
        Polygon2=np.array(
            [
                [0, 0],
                [6, 0],
                [6, 8],
                [0, 8],
            ],
            dtype=np.float32,
        ),
        Expected=np.array(
            [
                [0, 0],
                [6, 0],
                [6, 8],
                [0, 8],
            ],
            dtype=np.float32,
        ),
        IoU=48 / 80,
    ),
]


@dataclass
class PolygonArea:
    Polygon: np.ndarray
    Area: np.float32

    def ragged(l):
        mask = tf.ragged.constant([[True] * d.Polygon.shape[0] for d in l])
        vertices = tf.ragged.constant([d.Polygon for d in l]).to_tensor()
        polygons = tf.where(
            mask.to_tensor()[..., None], vertices, vertices[:, 0, :][:, None, :]
        )
        return polygons, tf.constant([d.Area for d in l])


POLYGON_AREA = [
    PolygonArea(
        Polygon=np.array(
            [
                [0, 0],
                [0, 1],
                [1, 1],
                [1, 0],
            ],
            dtype=np.float32,
        ),
        Area=1,
    ),
    PolygonArea(
        Polygon=np.array(
            [
                [0, 0],
                [0, 0],
                [0, 1],
                [0, 1],
                [1, 1],
                [1, 1],
                [1, 0],
                [1, 0],
            ],
            dtype=np.float32,
        ),
        Area=1,
    ),
]
